package awais.instagrabber.utils;

import android.Manifest;
import android.annotation.SuppressLint;
import android.app.Activity;
import android.app.NotificationManager;
import android.content.ClipData;
import android.content.ClipboardManager;
import android.content.ContentResolver;
import android.content.Context;
import android.content.pm.PackageManager;
import android.content.res.Resources;
import android.net.Uri;
import android.os.AsyncTask;
import android.os.Environment;
import android.os.Handler;
import android.os.HandlerThread;
import android.util.DisplayMetrics;
import android.util.Log;
import android.webkit.CookieManager;
import android.webkit.MimeTypeMap;
import android.widget.Toast;

import androidx.annotation.NonNull;
import androidx.annotation.Nullable;
import androidx.core.app.ActivityCompat;
import androidx.core.content.ContextCompat;

import com.squareup.picasso.Picasso;

import org.json.JSONArray;
import org.json.JSONObject;

import java.io.BufferedReader;
import java.io.File;
import java.io.InputStreamReader;
import java.net.CookiePolicy;
import java.net.CookieStore;
import java.net.HttpCookie;
import java.net.HttpURLConnection;
import java.net.URI;
import java.net.URISyntaxException;

import awais.instagrabber.BuildConfig;
import awais.instagrabber.activities.Main;
import awais.instagrabber.asyncs.DownloadAsync;
import awais.instagrabber.asyncs.PostFetcher;
import awais.instagrabber.interfaces.FetchListener;
import awais.instagrabber.models.ClipModel;
import awais.instagrabber.models.PostModel;
import awais.instagrabber.models.ViewerPostModel;
import awais.instagrabber.utils.dialogs.SettingConstants;

public class Utils {
    @SuppressLint("StaticFieldLeak")
    public static final Picasso PICASSO = Picasso.get();
    public static final CookieManager COOKIE_MANAGER = CookieManager.getInstance();
    public static final String[] PERMS = new String[]{Manifest.permission.WRITE_EXTERNAL_STORAGE};
    public static final java.net.CookieManager NET_COOKIE_MANAGER = new java.net.CookieManager(null, CookiePolicy.ACCEPT_ALL);
    public static final MimeTypeMap mimeTypeMap = MimeTypeMap.getSingleton();
    public static final String CHANNEL_ID = "InstaGrabber", CHANNEL_NAME = "Instagrabber";
    public static boolean isChannelCreated = false;
    public static boolean isInstagramInstalled = false;
    public static NotificationManager notificationManager;
    public static ClipboardManager clipboardManager;
    private static String lastClipContent;
    static DisplayMetrics displayMetrics = Resources.getSystem().getDisplayMetrics();

    public static void startClipboardMonitor(final FetchListener<ClipModel> listener) {
        lastClipContent = getClipText();
        final HandlerThread handlerThread = new HandlerThread("instagrab_clip_monitor");
        handlerThread.start();
        final Handler handler = new Handler(handlerThread.getLooper());
        handler.post(new Runnable() {
            @Override
            public void run() {
                final String clipContent = getClipText();
                if (clipContent != null && !clipContent.equals(lastClipContent)) {
                    lastClipContent = clipContent;
                    final ClipModel clipModel = stripString(clipContent);
                    if (clipModel != null) listener.onResult(clipModel);
                }
                handler.postDelayed(this, 517);
            }
        });
    }

    public static void setupCookies(final String cookieRaw) {
        if (cookieRaw != null && !isEmpty(cookieRaw)) {
            final CookieStore cookieStore = NET_COOKIE_MANAGER.getCookieStore();
            try {
                final URI uri1 = new URI("https://instagram.com");
                final URI uri2 = new URI("https://instagram.com/");
                for (final String cookie : cookieRaw.split(";")) {
                    final String[] strings = cookie.split("=", 2);
                    final HttpCookie httpCookie = new HttpCookie(strings[0].trim(), strings[1].trim());
                    httpCookie.setDomain("instagram.com");
                    httpCookie.setPath("/");
                    httpCookie.setVersion(0);
                    cookieStore.add(uri1, httpCookie);
                    cookieStore.add(uri2, httpCookie);
                }
            } catch (final URISyntaxException e) {
                Log.e("AWAISKING_APP", "", e);
            }
        }
    }

    @Nullable
    public static ClipModel stripString(@NonNull String clipString) {
        final int wwwDel = clipString.contains("www.") ? 4 : 0;
        boolean isPost = false;
        if (clipString.contains("instagram.com/")) {
            clipString = clipString.substring(clipString.startsWith("https") ? 22 + wwwDel : 21 + wwwDel);

            final char firstChar = clipString.charAt(0);
            if ((firstChar == 'p' || firstChar == 'P') && clipString.charAt(1) == '/') {
                clipString = clipString.substring(2);
                isPost = true;
            }

            clipString = cleanString(clipString);
        } else if (clipString.contains("ig.me/u/")) {
            clipString = clipString.substring(clipString.startsWith("https") ? 16 + wwwDel : 15 + wwwDel);
            clipString = cleanString(clipString);
            isPost = false;
        } else return null;

        if (clipString.charAt(clipString.length() - 1) == '/')
            clipString = clipString.substring(0, clipString.length() - 1);

        return new ClipModel(isPost, clipString);
    }

    @NonNull
    public static String cleanString(@NonNull final String clipString) {
        final int queryIndex = clipString.indexOf('?');
        final int paramIndex = clipString.indexOf('#');
        int startIndex = -1;
        if (queryIndex > 0 && paramIndex > 0) {
            if (queryIndex < paramIndex) startIndex = queryIndex;
            else if (paramIndex < queryIndex) startIndex = paramIndex;
        } else if (queryIndex == -1 && paramIndex > 0) startIndex = paramIndex;
        else if (paramIndex == -1 && queryIndex > 0) startIndex = queryIndex;
        return startIndex != -1 ? clipString.substring(0, startIndex) : clipString;
    }

    @Nullable
    private static String getClipText() {
        String result = null;
        if (clipboardManager != null && clipboardManager.hasPrimaryClip()) {
            final ClipData clip = clipboardManager.getPrimaryClip();
            if (clip != null) {
                final CharSequence text = clip.getItemAt(0).getText();
                result = text != null ? text.toString() : null;
            }
        }
        return result;
    }

    @Nullable
    public static String getHighQualityPost(final JSONArray resources, final boolean isVideo) {
        try {
            final int resourcesLen = resources.length();

            final String[] sources = new String[resourcesLen];
            int lastResMain = 0, lastIndexMain = -1;
            int lastResBase = 0, lastIndexBase = -1;
            for (int i = 0; i < resourcesLen; ++i) {
                final JSONObject item = resources.getJSONObject(i);
                if (item != null && (!isVideo || item.has("profile"))) {
                    sources[i] = item.getString("src");
                    final int currRes = item.getInt("config_width") * item.getInt("config_height");

                    final String profile = isVideo ? item.getString("profile") : null;

                    if (!isVideo || "MAIN".equals(profile)) {
                        if (currRes > lastResMain) {
                            lastResMain = currRes;
                            lastIndexMain = i;
                        }
                    } else if ("BASELINE".equals(profile)) {
                        if (currRes > lastResBase) {
                            lastResBase = currRes;
                            lastIndexBase = i;
                        }
                    }
                }
            }

            if (lastIndexMain >= 0) return sources[lastIndexMain];
            else if (lastIndexBase >= 0) return sources[lastIndexBase];
        } catch (final Exception e) {
            Log.e("AWAISKING_APP", "", e);
        }
        return null;
    }

    public static String getHighQualityImage(final JSONObject resources) {
        String src = null;
        try {
            if (resources.has("display_resources")) src = getHighQualityPost(resources.getJSONArray("display_resources"), false);
            if (src == null) return resources.getString("display_url");
        } catch (final Exception e) {
            Log.e("AWAISKING_APP", "", e);
        }
        return src;
    }

    public static int convertDpToPx(int dp) {
        if (displayMetrics == null)
            displayMetrics = Resources.getSystem().getDisplayMetrics();
        return Math.round((dp * displayMetrics.densityDpi) / 160.0f);
    }

    public static void copyText(Context context, String string) {
        if (context != null) {
            if (clipboardManager == null)
                clipboardManager = (ClipboardManager) context.getSystemService(Context.CLIPBOARD_SERVICE);
            if (clipboardManager == null || !clipboardManager.hasPrimaryClip()) {
                Toast.makeText(context, "Error copying text", Toast.LENGTH_SHORT).show();
                return;
            }
        }
        if (clipboardManager != null) {
            clipboardManager.setPrimaryClip(ClipData.newPlainText(Utils.CHANNEL_NAME, string));
            if (context != null) Toast.makeText(context, "Copied to clipboard!", Toast.LENGTH_SHORT).show();
        }
    }

    @NonNull
    public static String readFromConnection(@NonNull HttpURLConnection conn) throws Exception {
        final BufferedReader br = new BufferedReader(new InputStreamReader(conn.getInputStream()));
        final StringBuilder sb = new StringBuilder();
        String line;
        while ((line = br.readLine()) != null) sb.append(line).append('\n');
        br.close();
        return sb.toString();
    }

    public static void batchDownload(@NonNull final Main main) {
        if (ContextCompat.checkSelfPermission(main, Utils.PERMS[0]) == PackageManager.PERMISSION_GRANTED)
            batchDownloadImpl(main);
        else
            ActivityCompat.requestPermissions(main, Utils.PERMS, 17);
    }

    public static void batchDownloadImpl(@NonNull final Main main) {
        final File[] dir = {new File(Environment.getExternalStorageDirectory(), "Download")};
        if (!BuildConfig.DEBUG && Main.sharedPreferences.getBoolean(SettingConstants.DOWNLOAD_USER_FOLDER, false))
            dir[0] = new File(dir[0], main.userQuery);

        if (!dir[0].exists() && !dir[0].mkdirs()) {
            Toast.makeText(main, "Error creating Download folder(s).", Toast.LENGTH_SHORT).show();
            return;
        }

        //final ArrayList<PostModel> failedToDownloadList = new ArrayList<>(); //todo?
        for (final PostModel selectedItem : main.selectedItems) {
            new PostFetcher(selectedItem.getShortCode(), result -> {
                if (result == null) {
                    main.mainHelper.deselectSelection(selectedItem);
                    //failedToDownloadList.add(selectedItem);
                    return;
                }

                final int resultsSize = result.length;
                for (int i = 0; i < resultsSize; i++) {
                    final ViewerPostModel model = result[i];
                    final String displayUrl = model.getDisplayUrl();
                    final int index = displayUrl.indexOf('?');
                    final File outFile = new File(dir[0], model.getPostId() + "_" + model.getTimestamp() +
                            (resultsSize > 1 ? ("_slide_" + (i + 1)) : "") + displayUrl.substring(index - 4, index));
                    new DownloadAsync(main, displayUrl, outFile, file -> {
                        main.mainHelper.deselectSelection(selectedItem);
                        //if (file == null) failedToDownloadList.add(selectedItem);
                    }).executeOnExecutor(AsyncTask.THREAD_POOL_EXECUTOR);
                }
            }).executeOnExecutor(AsyncTask.THREAD_POOL_EXECUTOR);
        }
    }

    public static boolean checkKey(final String key, final String username, final String name) {
        if (!Utils.isEmpty(key)) {
            final boolean hasUserName = username != null && username.toLowerCase().contains(key);
            if (!hasUserName && name != null) return name.toLowerCase().contains(key);
        }
        return true;
    }

    // taken from Arrays.toString()
    @NonNull
    public static String highlightIdsMerger(final String... strings) {
        if (strings != null) {
            int iMax = strings.length - 1;
            if (iMax != -1) {
                final StringBuilder builder = new StringBuilder();
                builder.append('[');
                for (int i = 0; ; i++) {
                    builder.append('"').append(strings[i]).append('"');
                    if (i == iMax) return builder.append(']').toString();
                    builder.append(',');
                }
            }

        }
        return "[]";
    }

    public static boolean isEmpty(String str) {
        if (str == null || "".equals(str) || "null".equals(str) || str.isEmpty()) return true;
        str = str.trim();
        return "".equals(str) || "null".equals(str) || str.isEmpty();
    }

    public static boolean isImage(final Uri itemUri, final ContentResolver contentResolver) {
        String mimeType;
        if (itemUri == null) return false;
        final String scheme = itemUri.getScheme();
        if (isEmpty(scheme))
            mimeType = mimeTypeMap.getMimeTypeFromExtension(MimeTypeMap.getFileExtensionFromUrl(itemUri.toString()).toLowerCase());
        else mimeType = scheme.equals(ContentResolver.SCHEME_CONTENT) ? contentResolver.getType(itemUri)
                : mimeTypeMap.getMimeTypeFromExtension
                (MimeTypeMap.getFileExtensionFromUrl(itemUri.toString()).toLowerCase());

        if (isEmpty(mimeType)) return true;
        mimeType = mimeType.toLowerCase();
        return mimeType.startsWith("image");
    }

    @Nullable
    public static String getCookie(@Nullable final String webViewUrl) {
        int lastLongestCookieLength = 0;
        String mainCookie = null;

        String cookie;
        if (!Utils.isEmpty(webViewUrl)) {
            cookie = Utils.COOKIE_MANAGER.getCookie(webViewUrl);
            if (cookie != null) {
                final int cookieLen = cookie.length();
                if (cookieLen > lastLongestCookieLength) {
                    mainCookie = cookie;
                    lastLongestCookieLength = cookieLen;
                }
            }
        }
        cookie = Utils.COOKIE_MANAGER.getCookie("https://instagram.com");
        if (cookie != null) {
            final int cookieLen = cookie.length();
            if (cookieLen > lastLongestCookieLength) {
                mainCookie = cookie;
                lastLongestCookieLength = cookieLen;
            }
        }
        cookie = Utils.COOKIE_MANAGER.getCookie("https://instagram.com/");
        if (cookie != null) {
            final int cookieLen = cookie.length();
            if (cookieLen > lastLongestCookieLength) {
                mainCookie = cookie;
                lastLongestCookieLength = cookieLen;
            }
        }
        cookie = Utils.COOKIE_MANAGER.getCookie("http://instagram.com");
        if (cookie != null) {
            final int cookieLen = cookie.length();
            if (cookieLen > lastLongestCookieLength) {
                mainCookie = cookie;
                lastLongestCookieLength = cookieLen;
            }
        }
        cookie = Utils.COOKIE_MANAGER.getCookie("http://instagram.com/");
        if (cookie != null) {
            final int cookieLen = cookie.length();
            if (cookieLen > lastLongestCookieLength) {
                mainCookie = cookie;
                lastLongestCookieLength = cookieLen;
            }
        }
        cookie = Utils.COOKIE_MANAGER.getCookie("https://www.instagram.com");
        if (cookie != null) {
            final int cookieLen = cookie.length();
            if (cookieLen > lastLongestCookieLength) {
                mainCookie = cookie;
                lastLongestCookieLength = cookieLen;
            }
        }
        cookie = Utils.COOKIE_MANAGER.getCookie("https://www.instagram.com/");
        if (cookie != null) {
            final int cookieLen = cookie.length();
            if (cookieLen > lastLongestCookieLength) {
                mainCookie = cookie;
                lastLongestCookieLength = cookieLen;
            }
        }
        cookie = Utils.COOKIE_MANAGER.getCookie("http://www.instagram.com");
        if (cookie != null) {
            final int cookieLen = cookie.length();
            if (cookieLen > lastLongestCookieLength) {
                mainCookie = cookie;
                lastLongestCookieLength = cookieLen;
            }
        }
        cookie = Utils.COOKIE_MANAGER.getCookie("http://www.instagram.com/");
        if (cookie != null && cookie.length() > lastLongestCookieLength) mainCookie = cookie;

        return mainCookie;
    }

    public static void errorFinish(@NonNull final Activity activity) {
        Toast.makeText(activity, "Unkown error occurred!!", Toast.LENGTH_SHORT).show();
        activity.finish();
    }

    public static boolean isInstaInstalled(@NonNull final Activity activity) {
        final PackageManager packageManager = activity.getPackageManager();
        try {
            packageManager.getPackageInfo("com.instagram.android", 0);
            return true;
        } catch (final Exception e) {
            try {
                return packageManager.getApplicationInfo("com.instagram.android", 0).enabled;
            } catch (final Exception e1) {
                return false;
            }
        }
    }
}